/* This file is part of smap.
   Copyright (C) 2006, 2007, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#include "smap.h"
#ifdef HAVE_GETOPT_H
# include <getopt.h>
#endif

#ifndef SYSCONFDIR
# define SYSCONFDIR "/etc"
#endif

char *config_file = SYSCONFDIR "/smap.conf";
int foreground;
unsigned idle_timeout;
int inetd_mode;
int lint_mode;
char *pidfile;

int
read_delim(FILE *fp, int delim, char *buf, size_t bufsize)
{
	int c;
	int len = 0;

	alarm(idle_timeout);
	bufsize--;
	while (len < bufsize && ((c = fgetc (fp))) != delim) {
		if (c == EOF) {
			if (len) {
				if (debug_lev_p(DBG_SMAP, 10))
					smap_ostream_printf(smap_debug_str,
							    "recv: %.*s:\n",
							    len, buf);
				smap_error("unexpected end of file on input");
				exit(1);
			}
			return 0;
		}
		buf[len++] = c;
	}
	alarm (0);
	buf[len] = 0;
	if (debug_lev_p(DBG_SMAP, 10))
		smap_ostream_printf(smap_debug_str, "recv: %s:", buf);
	return len;
}

void
smap_query(smap_ostream_t ostr, const char *map, const char *key,
	   struct sockaddr const *sa, socklen_t socklen)
{
	smap_ostream_printf(ostr, "OK %s %s\n", map, key);
}

int
smap_session_server(const char *id, int fd,
		    struct sockaddr const *sa, socklen_t salen,
		    void *server_data, void *srvman_data)
{
	FILE *in, *out;
	char lbuf[80];
	char *buf = NULL;
	size_t bufsize = 0;
	size_t len;
	char *key, *p;
	smap_ostream_t ostr;
	
	/* FIXME: priv. setup */
	in = fdopen(fd, "r");
	setvbuf(in, NULL, _IONBF, 0);
	out = fdopen(fd, "w");
	setvbuf(out, NULL, _IONBF, 0);

	ostr = smap_map_ostream_create(out);
	
	/* Read input: */
	while (read_delim(in, ':', lbuf, sizeof lbuf)) {
		len = strtoul(lbuf, &p, 10);
		if (*p != 0) {
			smap_error("protocol error: expected packet length,"
				   " but found %s",
				   buf);
			exit(1);
		}

		if (len > bufsize) {
			bufsize = len + 1;
			buf = erealloc(buf, bufsize);
		}
		
		if (fread(buf, 1, len, in) != len) {
			smap_error("protocol error: short read");
			exit(1);
		}

		buf[len] = 0;
  
		if (debug_lev_p(DBG_SMAP, 10))
			smap_ostream_printf(smap_debug_str, "%s,\n", buf);

		if (getc(in) != ',') {
			smap_error("protocol error: missing"
				   " terminating comma");
			exit(1);
		}

		key = strchr(buf, ' ');
		if (!key) {
			smap_error("protocol error: missing map name");
			exit(1);
		}
		*key++ = 0;

		len = strlen(key);
		if (key[len-1] == '*') /* FIXME: why? */
			key[len-1] = 0;

		if (smap_trace_option)
			smap_ostream_printf(smap_trace_str,
					    "%s %s => ",
					    buf, key);
		smap_query(ostr, buf, key, sa, salen);
	}
	/* Cleanup and exit */
	free(buf);
	smap_ostream_free(ostr);
	fclose(in);
	fclose(out);
	return 0;
}


static int restart;         /* Set to 1 if restart is requested */

static RETSIGTYPE
sig_stop(int sig)
{
	smap_srvman_stop();
}

static RETSIGTYPE
sig_restart(int sig)
{
	restart = 1;
	smap_srvman_stop();
}



void
smap_daemon(int argc, char **argv)
{
	if (!foreground) {
		if (daemon(0, 0) < 0) {
			smap_error("failed to become a daemon");
			exit(EX_UNAVAILABLE);
		}
	}

	if (pidfile) {
		FILE *fp = fopen(pidfile, "w");
		if (!fp) {
			smap_error("cannot open pidfile %s for writing: %s",
				   pidfile, strerror(errno));
		} else {
			fprintf(fp, "%lu\n", (unsigned long) getpid());
			fclose(fp);
		}
	}

	signal(SIGTERM, sig_stop);
	signal(SIGQUIT, sig_stop);
        if (argv[0][0] != '/') {
		smap_error(_("smap started "
			     "using relative file name; "
			     "restart (SIGHUP) will not work"));
		signal(SIGHUP, sig_stop);
	} else
		signal(SIGHUP, sig_restart);
	signal(SIGINT, sig_stop);
	
        if (smap_srvman_open()) {
		smap_error("no servers configured; exiting");
		exit(EX_UNAVAILABLE);
	}
	smap_srvman_run(NULL);
	smap_srvman_shutdown();
	smap_srvman_free();
	if (pidfile && unlink(pidfile))
		smap_error("failed to remove pidfile %s: %s",
			   pidfile, strerror(errno));
	if (restart) {
		char *s;
		
		smap_error(_("smap restarting"));
		close_fds_above(1);
		execv(argv[0], argv);
		smap_log_init();
		smap_error(_("cannot restart: %s"), strerror(errno));
	} else
		smap_error(_("smap terminating"));
	exit(0);
}


#define CFG_GETNUM(str, val)						\
	do {								\
		char *__p;						\
		unsigned long __n = strtoul(str, &__p, 0);		\
		if (*__p) {						\
			smap_error("%s:%lu: invalid number (near %s)",	\
				   cfg_file_name, cfg_line, __p);	\
			return 1;					\
		}							\
		if ((val = __n) != __n) {				\
			smap_error("%s:%lu: number out of range: %s",	\
				   cfg_file_name, cfg_line, __p);	\
			return 1;					\
		}							\
	}  while (0)

static int
bool_invert(struct cfg_kw *kw, int wordc, char **wordv)
{
	*kw->ival = !*kw->ival;
	return 0;
}

static int
set_syslog_facility(struct cfg_kw *kw, int wordc, char **wordv)
{
	if (cfg_chkargc(wordc, 2, 2))
		return 1;
	
	if (syslog_fname_to_n(wordv[1], kw->ival)) {
		smap_error("%s:%u: unrecognized facility: %s", wordv[1],
			   cfg_file_name, cfg_line);
		return 1;
	}
	return 0;
}

static int
cfg_debug(struct cfg_kw *kw, int wordc, char **wordv)
{
	int i;
	
	if (cfg_chkargc(wordc, 2, 2))
		return 1;
	for (i = 1; i < wordc; i++) {
		if (debug_set(wordv[i])) {
			smap_error("%s:%lu: invalid debug specification: %s",
				   cfg_file_name, cfg_line, wordv[i]);
			return 1;
		}
	}
	return 0;
}

static int
cfg_backlog(struct cfg_kw *kw, int wordc, char **wordv)
{
	int n;
	
	if (cfg_chkargc(wordc, 2, 3))
		return 1;
	if (wordc == 3) {
		smap_server_t srv = smap_srvman_find_server(wordv[1]);
		if (!srv) {
			smap_error("%s:%lu: no such server: %s",
				   cfg_file_name, cfg_line, wordv[1]);
			return 1;
		}
		CFG_GETNUM(wordv[2], n);
		smap_server_set_backlog(srv, n);
	} else {
		CFG_GETNUM(wordv[1], n);
		srvman_param.backlog = n;
	}
	return 0;
}

static int
cfg_server_flag(struct cfg_kw *kw, int wordc, char **wordv, int flag,
		int *pglobal)
{
	int n;
	
	if (cfg_chkargc(wordc, 2, 3))
		return 1;
	if (wordc == 3) {
		smap_server_t srv = smap_srvman_find_server(wordv[1]);
		if (!srv) {
			smap_error("%s:%lu: no such server: %s",
				   cfg_file_name, cfg_line, wordv[1]);
			return 1;
		}
		if (cfg_parse_bool(wordv[2], &n))
			return 1;
		smap_server_set_flags(srv, flag,
				      n ? srvman_bitop_set : srvman_bitop_clr);
	} else {
		if (cfg_parse_bool(wordv[1], &n))
			return 1;
		*pglobal = n;
	}
	return 0;
}

static int
cfg_reuseaddr(struct cfg_kw *kw, int wordc, char **wordv)
{
	return cfg_server_flag(kw, wordc, wordv, SRV_KEEP_EXISTING,
			       &srvman_param.reuseaddr);
}

static int
cfg_single_process(struct cfg_kw *kw, int wordc, char **wordv)
{
	return cfg_server_flag(kw, wordc, wordv, SRV_SINGLE_PROCESS,
			       &srvman_param.single_process);
}

static int
cfg_max_children(struct cfg_kw *kw, int wordc, char **wordv)
{
	size_t n;
	
	if (cfg_chkargc(wordc, 2, 3))
		return 1;
	if (wordc == 3) {
		smap_server_t srv = smap_srvman_find_server(wordv[1]);
		if (!srv) {
			smap_error("%s:%lu: no such server: %s",
				   cfg_file_name, cfg_line, wordv[1]);
			return 1;
		}
		CFG_GETNUM(wordv[2], n);
		smap_server_set_max_children(srv, n);
	} else {
		CFG_GETNUM(wordv[1], n);
		srvman_param.max_children = n;
	}
	return 0;
}

static int
cfg_server(struct cfg_kw *kw, int wordc, char **wordv)
{
	size_t n;
	smap_server_t srv;
	
	if (cfg_chkargc(wordc, 3, 3))
		return 1;
	srv = smap_server_new(wordv[1], wordv[2], smap_session_server, 0);
	if (srv)
		smap_srvman_attach_server(srv);
	else
		return 1; /* FIXME: Error message */
	return 0;
}

static struct cfg_kw smap_kwtab[] = {
	/*   kw        type       ival   sval  aval  fun */
	{ "pidfile", KWT_STRING, NULL, &pidfile, NULL, NULL },
	{ "foreground", KWT_BOOL, &foreground, },
	{ "idle-timeout", KWT_UINT, &idle_timeout },
	{ "log-to-stderr", KWT_BOOL, &log_to_stderr, },
	{ "log-to-syslog", KWT_BOOL, &log_to_stderr, NULL, NULL, bool_invert },
	{ "log-tag", KWT_STRING, NULL, &log_tag },
	{ "log-facility", KWT_FUN, &log_facility, NULL, NULL,
	  set_syslog_facility },
	{ "debug", KWT_FUN, NULL, NULL, NULL, cfg_debug },

	/* Srvman global parameters: */
	{ "socket-umask", KWT_INT_O, &srvman_param.socket_umask },
	{ "shutdown-timeout", KWT_UINT,
	  (int*) &srvman_param.shutdown_timeout },

	/* Server declarations */
	{ "server", KWT_FUN, NULL, NULL, NULL, cfg_server },
	{ "backlog", KWT_FUN, NULL, NULL, NULL, cfg_backlog },
	{ "reuseaddr", KWT_FUN, NULL, NULL, NULL, cfg_reuseaddr },
	{ "max-children", KWT_FUN, NULL, NULL, NULL, cfg_max_children },
	{ "single-process", KWT_FUN, NULL, NULL, NULL, cfg_single_process },
	
	{ NULL }
};


#include "cmdline.c"

int
main(int argc, char **argv)
{
	set_program_name(argv[0]);
	smap_log_init();
	get_options(argc, argv);
	if (log_to_stderr == -1)
		log_to_stderr = foreground;
	smap_log_init();
	if (parse_config(config_file, smap_kwtab))
		exit(EX_CONFIG);
	if (lint_mode)
		exit(0);

	smap_daemon(argc, argv);
	exit(0);
}
